"""Internal utilities for class deserialization."""

import logging
from dataclasses import InitVar, dataclass
from typing import Any, List, Sequence, TypeVar, Union

import terramare.metadata
from terramare.errors import InternalError

from .. import type_utils
from ..core import InternalConstructor, InternalFactory
from ..data import Context
from ..metadata import MetadataCollection, with_
from ..types import Target

_log = logging.getLogger(__name__)


class _PseudoField:
    pass


_T_co = TypeVar("_T_co", covariant=True)


@dataclass(frozen=True)
class SingleField:
    constructor: InternalConstructor[Any]
    required: bool

    @staticmethod
    def from_parameter(
        metadata: MetadataCollection,
        factory: InternalFactory,
        target: Target[_T_co],
        parameter: "TypedParameter",
    ) -> "SingleField":
        with_fn = with_.read_field(metadata, target, parameter.name)
        if with_fn is None:
            parameter_type = parameter.type_hint
        else:
            parameter_type = with_fn()
        return SingleField(
            factory.create_constructor(
                parameter_type, frame=f'field "{parameter.name}"'
            ),
            parameter.is_required(),
        )


@dataclass(frozen=True)
class ContextField(_PseudoField):
    pass


NO_DEFAULT = object()


@dataclass(frozen=True)
class FromContextField(_PseudoField):
    key: str
    type_: Target[Any]
    _default: object

    @property
    def required(self) -> bool:
        return self._default == NO_DEFAULT

    @property
    def default(self) -> object:
        if self._default == NO_DEFAULT:  # pragma: no cover
            raise InternalError(
                f"attempted to retrieve default of required from_context parameter {self.key}"
            )
        return self._default

    @staticmethod
    def from_parameter(key: str, parameter: "TypedParameter") -> "FromContextField":
        return FromContextField(
            key,
            parameter.type_hint,
            NO_DEFAULT if parameter.is_required() else parameter.default,
        )


@dataclass(frozen=True)
class SkippedField(_PseudoField):
    default: object


PseudoField = Union[ContextField, FromContextField, SkippedField]

Field = Union[SingleField, PseudoField]


@dataclass(frozen=True)
class VariadicField:
    constructor: InternalConstructor[Any]


@dataclass(frozen=True)
class TypedParameter(type_utils.Parameter):
    type_hint: Target[Any]

    def is_context(self) -> bool:
        return self.type_hint is Context


def add_type_hints(
    type_: Target[Any], params: Sequence[type_utils.Parameter]
) -> List[TypedParameter]:
    type_hints = type_utils.get_type_hints(type_, params)
    return [
        TypedParameter(p.parameter, _strip_initvar(type_hints[p.name])) for p in params
    ]


def skip(
    metadata: MetadataCollection,
    factory: InternalFactory,
    target: Target[Any],
    parameter: type_utils.Parameter,
) -> bool:
    if not terramare.metadata.skip.read_field(metadata, target, parameter.name):
        return False
    if parameter.is_required():
        raise factory.make_error(f"cannot skip required parameter: {parameter.name}")
    return True


def handle_positional_pseudo_field(
    index: int, field: PseudoField, context: Context
) -> object:
    if isinstance(field, ContextField):
        _log.debug("Positional parameter %d using context: %s", index, context)
        return context
    elif isinstance(field, FromContextField):
        if field.key in context:
            value = context[field.key]
            _log.debug(
                "Positional parameter %d using context field '%s': %s",
                index,
                field.key,
                value,
            )
            return value
        elif field.required:  # pragma: no cover
            raise InternalError(f"missing required context: {field.key}")
        else:
            _log.debug(
                "Positional parameter %d using default value for context field '%s': %s",
                index,
                field.key,
                field.default,
            )
            return field.default
    elif isinstance(field, SkippedField):
        _log.debug(
            "Positional parameter %d using default value for skipped field: %s",
            index,
            field.default,
        )
        return field.default
    else:  # pragma: no cover
        raise InternalError(f"Unknown PseudoField variant ({type(field)}): {field}")


OMIT = object()


def handle_keyword_pseudo_field(
    name: str, field: PseudoField, context: Context
) -> object:
    if isinstance(field, ContextField):
        _log.debug("Keyword parameter '%s' using context: %s", name, context)
        return context
    elif isinstance(field, FromContextField):
        if field.key in context:
            value = context[field.key]
            _log.debug(
                "Keyword parameter '%s' using context field '%s': %s",
                name,
                field.key,
                value,
            )
            return value
        elif field.required:  # pragma: no cover
            raise InternalError(f"missing required context: {field.key}")
        else:
            _log.debug(
                "Omitting keyword parameter '%s' for optional context field '%s'",
                name,
                field.key,
            )
            return OMIT
    elif isinstance(field, SkippedField):
        _log.debug("Keyword parameter '%s' field skipped", name)
        return OMIT
    else:  # pragma: no cover
        raise InternalError(f"Unknown PseudoField variant ({type(field)}): {field}")


def _strip_initvar(target: Target[object]) -> Target[object]:
    # One or other of these branches is covered, depending on the Python
    # version.
    if isinstance(target, InitVar) and hasattr(target, "type"):  # pragma: no cover
        return target.type
    if target is InitVar:  # pragma: no cover
        raise type_utils.ParameterError(
            "'InitVar' is not supported before Python 3.8 as it does not "
            "contain the necessary type information - see "
            "https://bugs.python.org/issue33569."
        )
    return target



"""Primitive wrapper types which maintain context during construction."""

import collections.abc
import dataclasses
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from typing_extensions import final

from .errors import ConstructorError
from .pretty_printer import print_primitive, print_table, print_type_name
from .types import ArrayPrimitive, ObjectPrimitive, Primitive, Target


@dataclass(frozen=True)
class ConstructionConfig:
    coerce_strings: bool = False


@dataclass(frozen=True)
class Context(Mapping[str, object]):
    """
    Arbitrary user-provided construction context.

    The context is provided to :any:`terramare.structure` and is made available
    to classes during construction via parameters of this type.
    """

    EMPTY: ClassVar["Context"]

    _inner: Mapping[str, object] = field(default_factory=dict)

    def __getitem__(self, __key: str) -> object:
        return self._inner.__getitem__(__key)

    def __iter__(self) -> Iterator[str]:
        return self._inner.__iter__()

    def __len__(self) -> int:
        return self._inner.__len__()


Context.EMPTY = Context()

_Self = TypeVar("_Self", bound="Value")

_Index = Union[int, str]


@dataclass(frozen=True, order=False)
class Value:
    raw: Primitive
    config: ConstructionConfig
    context: Context
    _parent: Optional[Tuple[_Index, "Value"]] = field(compare=False)
    _stack: Sequence[Target[Any]] = field(default_factory=list, compare=False)

    @classmethod
    def new(
        cls: Type[_Self],
        data: Primitive,
        *,
        config: Optional[ConstructionConfig] = None,
        context: Context = Context.EMPTY,
    ) -> _Self:
        return cls(data, config or ConstructionConfig(), context, None)

    def is_array(self) -> bool:
        return isinstance(self.raw, list)

    def as_array(self) -> "Array":
        if not self.is_array():
            raise self.make_error("expected array")
        return Array(
            cast(ArrayPrimitive, self.raw),
            self.config,
            self.context,
            self._parent,
            self._stack,
        )

    def is_object(self) -> bool:
        return isinstance(self.raw, dict)

    def as_object(self) -> "Object":
        if not self.is_object():
            raise self.make_error("expected object")
        return Object(
            cast(ObjectPrimitive, self.raw),
            self.config,
            self.context,
            self._parent,
            self._stack,
        )

    def push_stack(self: _Self, frame: Target[Any]) -> _Self:
        return dataclasses.replace(self, _stack=[*self._stack, frame])

    def make_error(self, msg: str) -> ConstructorError:
        # We are accessing private members of other instances of Value.
        # These members are not for public use but are expected to be used in
        # this way.
        # pylint: disable=protected-access
        return ConstructorError(
            f"{self._get_path()}: {msg}",
            print_table(
                [("path", "expected structure", "actual data")]
                + [
                    (
                        v._get_path(),
                        "/".join(print_type_name(t) for t in v._stack),
                        print_primitive(v.raw),
                    )
                    for v in self._iter_parents()
                ]
            ),
            self._get_depth(),
        )

    def _get_path(self) -> str:
        # As above, we are accessing private members of other instances of
        # ValueView.
        # pylint: disable=protected-access
        path = [v._get_index() for v in self._iter_parents()]
        return "." + ".".join(
            _quote_stack_element(index) for index in path if index is not None
        )

    def _get_index(self) -> Optional[_Index]:
        if self._parent:
            return self._parent[0]
        return None

    def _get_depth(self) -> Tuple[int, int]:
        return (len(list(self._iter_parents())), len(self._stack))

    def _iter_parents(self) -> Iterator["Value"]:
        # As above, we are accessing private members of other instances of
        # ValueView.
        # pylint: disable=protected-access
        def inner() -> Iterator["Value"]:
            current: Value = self
            while current._parent:
                yield current
                _, current = current._parent
            yield current

        return reversed(list(inner()))


# See https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime.  # noqa
if TYPE_CHECKING:  # pragma: no cover
    _BaseArray = collections.abc.Sequence[Value]
else:
    _BaseArray = collections.abc.Sequence


@final
@dataclass(frozen=True)
class Array(Value, _BaseArray):
    raw: ArrayPrimitive

    @overload
    def __getitem__(self, index: int) -> Value:
        ...  # pragma: no cover

    @overload
    def __getitem__(self, index: slice) -> List[Value]:
        ...  # pragma: no cover

    def __getitem__(self, index: Union[int, slice]) -> Union[Value, List[Value]]:
        def wrap(i: int) -> Value:
            return Value(self.raw[i], self.config, self.context, (i, self))

        if isinstance(index, int):
            return wrap(index)
        return [wrap(i) for i in range(0, len(self))[index]]

    def __len__(self) -> int:
        return len(self.raw)

    def as_array(self) -> "Array":  # pragma: no cover
        return self


# See https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime.  # noqa
if TYPE_CHECKING:  # pragma: no cover
    _BaseObject = collections.abc.Mapping[str, Value]
else:
    _BaseObject = collections.abc.Mapping


@final
@dataclass(frozen=True)
class Object(Value, _BaseObject):
    raw: ObjectPrimitive
    _hidden_keys: Set[str] = field(default_factory=set)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Object):
            return NotImplemented  # pragma: no cover
        if set(self) != set(other):
            return False
        return all(self[k] == other[k] for k in self)

    def __getitem__(self, key: str) -> Value:
        if key not in self.raw or key in self._hidden_keys:
            raise KeyError(key)
        return Value(self.raw[key], self.config, self.context, (key, self))

    def __iter__(self) -> Iterator[str]:
        return (k for k in self.raw if k not in self._hidden_keys)

    def __len__(self) -> int:
        return len(self.raw) - len(self._hidden_keys)

    def pop(self, key: str) -> Tuple["Object", Value]:
        if key not in self.raw:
            raise KeyError(key)
        return (
            dataclasses.replace(self, _hidden_keys={*self._hidden_keys, key}),
            self[key],
        )

    def is_object(self) -> bool:
        return True

    def as_object(self) -> "Object":
        return self


def _quote_stack_element(frame: Union[int, str]) -> str:
    if isinstance(frame, int):
        return str(frame)
    return frame.replace("\\", "\\\\").replace('"', '\\"')



"""Terramare-specific metadata for attrs classes."""

import abc
import collections
import dataclasses
import logging
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import final

from .core import ConstructorCore, FactoryCore, InternalFactory
from .errors import InternalError, TerramareError, UnsupportedTargetTypeError
from .pretty_printer import LoggableTarget
from .safe_mapping import SafeMapping
from .types import Lazy, Target

METADATA_KEY = f"{__name__}.dataclass_metadata"
_METADATA_FIELD = "__terramare_terramare__"

_log = logging.getLogger(__name__)

_T_co = TypeVar("_T_co", covariant=True)
_T_in = TypeVar("_T_in")


@dataclass(frozen=True)
class MetadataError(TerramareError):
    message: str


_Metadata = TypeVar("_Metadata")
_Target = TypeVar("_Target", bound=Target[Any])


def write(target: _Target, key: str, value: Any) -> _Target:
    metadata = _read_from_target(target)
    if not metadata:
        metadata = {}
        setattr(target, _METADATA_FIELD, (metadata, target))
    metadata[key] = value
    return target


TargetMetadata = Sequence["Metadata[Any]"]


@dataclass(frozen=True)
class MetadataCollection:
    _metadata: SafeMapping[Target[Any], Mapping[str, Any]] = dataclasses.field(
        default_factory=SafeMapping
    )

    @classmethod
    def new(
        cls, metadata: Mapping[Target[Any], TargetMetadata]
    ) -> "MetadataCollection":
        return cls(
            SafeMapping(
                [
                    (
                        target,
                        {element.KEY: element.value for element in target_metadata},
                    )
                    for target, target_metadata in metadata.items()
                ]
            )
        )

    def read(
        self,
        target: Target[Any],
        key: str,
        default: _Metadata,
        *,
        field: Optional[str] = None,
    ) -> _Metadata:
        if field is not None:
            return self._read_field_metadata(target, key, default, field)
        return self._read_target_metadata(target, key, default)

    def _read_field_metadata(
        self, target: Target[Any], key: str, default: _Metadata, field: str
    ) -> _Metadata:
        field_metadata = fields.read(self, target).get(field, {})
        if key in field_metadata:
            value = cast(_Metadata, field_metadata[key].value)
            _log.debug(
                "Read metadata key '%s' for field '%s' of target %s "
                "from fields metadata: %s",
                key,
                field,
                LoggableTarget(target),
                value,
            )
            return value

        dataclass_metadata = _read_from_dataclass_field(target, field)
        if key in dataclass_metadata:
            value = cast(_Metadata, dataclass_metadata[key].value)
            _log.debug(
                "Read metadata key '%s' for field '%s' of target %s "
                "from dataclass field metadata: %s",
                key,
                field,
                LoggableTarget(target),
                value,
            )
            return value

        return default

    def _read_target_metadata(
        self, target: Target[Any], key: str, default: _Metadata
    ) -> _Metadata:
        if target in self._metadata:
            target_metadata = self._metadata[target]
            if key in target_metadata:
                value = cast(_Metadata, target_metadata[key])
                _log.debug(
                    "Read metadata key '%s' for target %s from metadata collection: %s",
                    key,
                    LoggableTarget(target),
                    value,
                )
                return value

        target_metadata = _read_from_target(target)
        if key in target_metadata:
            value = cast(_Metadata, target_metadata[key])
            _log.debug(
                "Read metadata key '%s' for target %s from target metadata: %s",
                key,
                LoggableTarget(target),
                value,
            )
            return value

        return default


class Metadata(Generic[_Metadata], abc.ABC):
    KEY: ClassVar[str]
    DEFAULT: ClassVar[_Metadata]

    @classmethod
    def read(cls, metadata: MetadataCollection, target: _Target) -> _Metadata:
        return metadata.read(target, cls.KEY, cls.DEFAULT)

    @property
    @abc.abstractmethod
    def value(self) -> _Metadata:
        ...  # pragma: no cover

    def __call__(self, target: _Target) -> _Target:
        return write(target, self.KEY, self.value)


# See https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime.  # noqa
if TYPE_CHECKING:  # pragma: no cover
    _BaseFieldMetadata = collections.abc.Mapping[str, Any]
else:
    _BaseFieldMetadata = collections.abc.Mapping


class FieldMetadata(Generic[_Metadata], abc.ABC, _BaseFieldMetadata):
    KEY: ClassVar[str]
    DEFAULT: ClassVar[_Metadata]

    @classmethod
    def read_field(
        cls, metadata: MetadataCollection, target: _Target, field: str
    ) -> _Metadata:
        return metadata.read(target, cls.KEY, cls.DEFAULT, field=field)

    @property
    @abc.abstractmethod
    def value(self) -> _Metadata:
        ...  # pragma: no cover

    def __getitem__(self, key: str) -> Any:
        return self._as_dict()[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._as_dict())

    def __len__(self) -> int:
        return len(self._as_dict())

    def _as_dict(self) -> Mapping[str, Any]:
        return {METADATA_KEY: self}


_FieldsMetadata = Mapping[str, Mapping[str, FieldMetadata[object]]]


@dataclass(frozen=True)
class fields(Metadata[_FieldsMetadata]):
    KEY: ClassVar[str] = f"{__name__}.fields"
    DEFAULT: ClassVar[_FieldsMetadata]

    _value: _FieldsMetadata

    def __init__(
        self,
        fields_metadata: Mapping[
            str, Union[FieldMetadata[Any], Sequence[FieldMetadata[Any]]]
        ],
    ) -> None:
        def normalize(
            field_metadata: Union[FieldMetadata[Any], Sequence[FieldMetadata[Any]]]
        ) -> Sequence[FieldMetadata[object]]:
            if isinstance(field_metadata, FieldMetadata):  # pragma: no branch TODO
                field_metadata = [field_metadata]
            return field_metadata

        value: _FieldsMetadata = {
            field: {element.KEY: element for element in normalize(field_metadata)}
            for field, field_metadata in fields_metadata.items()
        }
        object.__setattr__(self, "_value", value)

    @property
    def value(self) -> _FieldsMetadata:
        return self._value


fields.DEFAULT = {}


class _skip(FieldMetadata[bool]):
    KEY: ClassVar[str] = f"{__name__}._skip"
    DEFAULT: ClassVar[bool] = False

    @property
    def value(self) -> bool:
        return True


skip = _skip()


@dataclass(frozen=True)
class from_context(FieldMetadata[Optional[str]]):
    KEY: ClassVar[str] = f"{__name__}.from_context"
    DEFAULT: ClassVar[Optional[str]] = None

    _value: str

    @property
    def value(self) -> str:
        return self._value


_With = Lazy[Callable[..., Any]]


@dataclass(frozen=True)
class with_(Metadata[Optional[_With]], FieldMetadata[Optional[_With]]):
    """Construct the decorated class using the given function."""

    KEY: ClassVar[str] = f"{__name__}.with_"
    DEFAULT: ClassVar[Optional[_With]] = None

    _value: _With

    @property
    def value(self) -> _With:
        return self._value


@final
@dataclass(frozen=True)
class MetadataFactory(FactoryCore):
    _metadata: MetadataCollection

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        _with = with_.read(self._metadata, target)
        if _with:
            return factory.create_constructor(_with(), frame="metadata 'with'").unwrap()
        raise UnsupportedTargetTypeError


def _read_from_target(target: Target[_T_in]) -> Dict[str, Any]:
    if not hasattr(target, _METADATA_FIELD):
        return {}
    attr = getattr(target, _METADATA_FIELD)
    if not (isinstance(attr, tuple) and isinstance(attr[0], dict)):
        raise MetadataError("unexpected type metadata")
    meta, t = attr
    # Ensure that the metadata was applied to `t` itself, rather than a base
    # class of `t`.
    # This is important because the base class may define metadata that is
    # not appropriate for derived classes, such as tagged polymorphism.
    if target != t:
        return {}
    return meta


_DATACLASS_FIELDS = "__dataclass_fields__"


def _read_from_dataclass_field(
    target: Target[Any], field: str
) -> Mapping[str, FieldMetadata[Any]]:
    if not (dataclasses.is_dataclass(target) and isinstance(target, type)):
        return {}
    if not hasattr(target, _DATACLASS_FIELDS):  # pragma: no cover
        raise InternalError(f"dataclass missing '{_DATACLASS_FIELDS}' field")
    field_data: Optional[dataclasses.Field[Any]] = getattr(
        target, _DATACLASS_FIELDS
    ).get(field)
    if field_data is None:
        return {}
    if METADATA_KEY not in field_data.metadata:
        return {}
    metadata = field_data.metadata[METADATA_KEY]

    if isinstance(metadata, FieldMetadata):
        return {metadata.KEY: metadata}
    if not (
        isinstance(metadata, list)
        and all(isinstance(v, FieldMetadata) for v in metadata)
    ):
        raise MetadataError("unexpected type metadata")
    return {v.KEY: v for v in metadata}



"""Deserializer to construct an instance of a class type from a value."""

import logging
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Mapping, Optional, Sequence, TypeVar

from .. import type_utils
from ..core import ConstructorCore, InternalConstructor, InternalFactory
from ..data import Value
from ..errors import InternalError
from ..metadata import MetadataCollection, from_context
from ..pretty_printer import LoggableArguments, LoggableTarget
from ..types import Target
from .utils import (
    OMIT,
    ContextField,
    Field,
    FromContextField,
    PseudoField,
    SingleField,
    SkippedField,
    TypedParameter,
    add_type_hints,
    handle_keyword_pseudo_field,
    handle_positional_pseudo_field,
    skip,
)

_log = logging.getLogger(__name__)

_T_co = TypeVar("_T_co", covariant=True)


@dataclass(frozen=True)
class FromValueConstructor(ConstructorCore[_T_co]):
    _positional: Sequence["Field"]
    _keyword: Mapping[str, "PseudoField"]
    # Variadic parameters are not used by this constructor.

    _target: Target[_T_co]

    def __post_init__(self) -> None:
        values = [f for f in self._positional if isinstance(f, SingleField)]
        if len(values) != 1:  # pragma: no cover
            raise InternalError(f"expected exactly one value field: {values}")

    @staticmethod
    def new(
        metadata: MetadataCollection,
        factory: InternalFactory,
        target: Target[_T_co],
    ) -> InternalConstructor[_T_co]:
        parameters = add_type_hints(target, type_utils.get_parameters(target))

        positional: List[Field] = []

        def have_value() -> bool:
            return any(isinstance(f, SingleField) for f in positional)

        index = 0
        while parameters and parameters[0].is_positional():
            parameter = parameters.pop(0)
            positional_field = FromValueConstructor._new_positional_field(
                metadata, factory, target, parameter, have_value()
            )
            if positional_field:
                _log.debug(
                    "Created field constructor for positional parameter '%s' (%d): %s",
                    parameter.name,
                    index,
                    positional_field,
                )
                positional.append(positional_field)
            else:
                _log.debug(
                    "No constructor created for positional parameter '%s' (%d): %s",
                    parameter.name,
                    index,
                    positional_field,
                )
            index += 1

        if not have_value():
            raise factory.make_error(
                "cannot create from-value constructor for "
                "target with no positional parameters"
            )

        positional_parameters = [p for p in parameters if p.is_positional()]
        if positional_parameters:  # pragma: no cover
            raise InternalError(
                f"unexpected positional parameters remaining: {positional_parameters}"
            )

        keyword: Dict[str, PseudoField] = {}
        while parameters:
            parameter = parameters.pop(0)
            keyword_field = FromValueConstructor._new_keyword_field(
                metadata, factory, target, parameter
            )
            if keyword_field:
                _log.debug(
                    "Created field constructor for keyword parameter '%s': %s",
                    parameter.name,
                    keyword_field,
                )
                keyword[parameter.name] = keyword_field
            else:
                _log.debug(
                    "No constructor created for keyword parameter '%s': %s",
                    parameter.name,
                    keyword_field,
                )

        if parameters:  # pragma: no cover
            raise InternalError(f"unexpected parameters remaining: {parameters}")

        return factory.wrap(
            target,
            FromValueConstructor(
                _positional=positional, _keyword=keyword, _target=target
            ),
        )

    @staticmethod
    def _new_positional_field(
        metadata: MetadataCollection,
        factory: InternalFactory,
        target: Target[_T_co],
        parameter: TypedParameter,
        have_value: bool,
    ) -> Optional[Field]:
        from_context_ = from_context.read_field(metadata, target, parameter.name)
        if skip(metadata, factory, target, parameter):
            return SkippedField(parameter.default)
        elif from_context_ is not None:
            return FromContextField.from_parameter(from_context_, parameter)
        elif parameter.is_context():
            if parameter.is_variadic():
                # The _FromValueConstructor doesn't use variadic context
                # parameters.
                return None
            return ContextField()
        elif not have_value:
            return SingleField(
                factory.create_constructor(
                    parameter.type_hint, frame=f'field "{parameter.name}"'
                ),
                parameter.is_required(),
            )
        elif parameter.is_required():
            raise factory.make_error(
                "cannot create from-value constructor for target "
                "with more than one required parameter"
            )
        # Covered, despite what 'coverage' says.
        else:  # pragma: no cover
            # Tail optional parameter; omit.
            return None

    @staticmethod
    def _new_keyword_field(
        metadata: MetadataCollection,
        factory: InternalFactory,
        target: Target[_T_co],
        parameter: TypedParameter,
    ) -> Optional[PseudoField]:
        from_context_ = from_context.read_field(metadata, target, parameter.name)
        if skip(metadata, factory, target, parameter):
            return SkippedField(parameter.default)
        elif from_context_ is not None:
            return FromContextField.from_parameter(from_context_, parameter)
        elif parameter.is_variadic():
            # The _FromValueConstructor doesn't use variadic parameters.
            return None
        elif parameter.is_context():
            return ContextField()
        elif parameter.is_required():
            raise factory.make_error(
                "cannot create from-value constructor for "
                "target with required keyword-only parameter"
            )
        else:
            # Keyword parameter is optional; omit it.
            return None

    def __call__(self, data: Value) -> _T_co:
        positional: List[object] = []
        for index, field in enumerate(self._positional):
            if isinstance(field, SingleField):
                value = field.constructor(data)
                _log.debug("Positional parameter %d using value: %s", index, value)
                positional.append(value)
            else:
                positional.append(
                    handle_positional_pseudo_field(index, field, data.context)
                )

        keyword: Dict[str, object] = {}
        for name, field in self._keyword.items():
            value = handle_keyword_pseudo_field(name, field, data.context)
            if value is not OMIT:
                keyword[name] = value

        _log.debug(
            "Instantiating %s with argument: %s",
            LoggableTarget(self._target),
            LoggableArguments(positional, keyword),
        )
        return self._target(*positional, **keyword)  # type: ignore[call-arg]

    def get_children(self) -> Iterator[InternalConstructor[Any]]:
        for field in self._positional:
            if isinstance(field, SingleField):
                yield field.constructor



"""Class-specific metadata."""

import abc
import enum
from dataclasses import dataclass
from typing import (
    AbstractSet,
    Any,
    ClassVar,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Final, Protocol

from ..metadata import Metadata
from ..types import Target

_Target = TypeVar("_Target", bound=Target[Any])


class _UnknownFieldConfigBase(abc.ABC):
    @abc.abstractmethod
    def ignore_element(self) -> bool:
        ...  # pragma: no cover

    @abc.abstractmethod
    def ignore_field(self, __name: str) -> bool:
        ...  # pragma: no cover


@dataclass(frozen=True)
class _AllowUnknownFields(_UnknownFieldConfigBase):
    def ignore_element(self) -> bool:
        return True

    def ignore_field(self, _: str) -> bool:
        return True


class _ExcludeIf(Protocol):
    def __call__(self, __field_name: str) -> bool:
        ...  # pragma: no cover


@dataclass(frozen=True)
class _DisallowUnknownFields(_UnknownFieldConfigBase):
    exclude_if: Optional[_ExcludeIf] = None

    def ignore_element(self) -> bool:
        return False

    def ignore_field(self, name: str) -> bool:
        if self.exclude_if is None:
            return False
        return self.exclude_if(name)


UnknownFieldConfig = Union[_AllowUnknownFields, _DisallowUnknownFields]
_DISALLOW_UNKNOWN_FIELDS_KEY = f"{__name__}.disallow_unknown_fields"


@dataclass(frozen=True)
class _disallow_unknown_fields(Metadata[UnknownFieldConfig]):

    KEY: ClassVar[str] = _DISALLOW_UNKNOWN_FIELDS_KEY
    DEFAULT: ClassVar[UnknownFieldConfig] = _AllowUnknownFields()

    @dataclass(frozen=True)
    class inner(Metadata[UnknownFieldConfig]):
        KEY: ClassVar[str] = _DISALLOW_UNKNOWN_FIELDS_KEY
        DEFAULT: ClassVar[UnknownFieldConfig] = _AllowUnknownFields()

        _value: UnknownFieldConfig

        @property
        def value(self) -> UnknownFieldConfig:
            return self._value

    @property
    def value(self) -> UnknownFieldConfig:
        return _DisallowUnknownFields()

    @overload
    def __call__(self, *, excluding: Set[str]) -> "_disallow_unknown_fields.inner":
        ...  # pragma: no cover

    @overload
    def __call__(self, *, exclude_if: _ExcludeIf) -> "_disallow_unknown_fields.inner":
        ...  # pragma: no cover

    @overload
    def __call__(self, target: _Target) -> _Target:
        ...  # pragma: no cover

    def __call__(
        self,
        target: Optional[_Target] = None,
        excluding: Optional[Set[str]] = None,
        exclude_if: Optional[_ExcludeIf] = None,
    ) -> Union["_disallow_unknown_fields.inner", _Target]:
        if excluding is not None:
            # First overload
            return _disallow_unknown_fields.inner(
                _DisallowUnknownFields(exclude_if=lambda s: s in excluding)
            )
        if exclude_if is not None:
            # Second overload
            return _disallow_unknown_fields.inner(
                _DisallowUnknownFields(exclude_if=exclude_if)
            )
        # Third overload
        assert target is not None
        return _disallow_unknown_fields.inner(self.value)(target)


#: Raise :exc:`terramare.errors.ConstructorError` if there are fields present in
#: the primitive that are not defined by the decorated class.
disallow_unknown_fields = _disallow_unknown_fields()


@dataclass(init=False)
class handle_exception_types(Metadata[Tuple[Type[Exception], ...]]):
    """
    Handle the listed exception types when constructing an instance of the decorated class.

    An exception of one of these types will be caught and re-raised as a
    :exc:`terramare.errors.ValidationError` including the full construction
    context.

    >>> import attr
    >>> import terramare
    >>>
    >>> @terramare.handle_exception_types(ValueError)
    ... @attr.s(auto_attribs=True)
    ... class User:
    ...     id: int = attr.ib()
    ...     name: str
    ...
    ...     @id.validator
    ...     def positive(self, _, value):
    ...         if not value > 0:
    ...             raise ValueError("id must be positive!")
    >>>
    >>> terramare.structure({"id": 0, "name": "Alice"}, into=User)
    Traceback (most recent call last):
    ...
    terramare.errors.ConstructorError: .: id must be positive!
    ...
    """

    KEY: ClassVar[str] = f"{__name__}.handle_exception_types"
    DEFAULT: ClassVar[Tuple[Type[Exception], ...]] = ()

    _value: Tuple[Type[Exception], ...]

    @property
    def value(self) -> Tuple[Type[Exception], ...]:
        return self._value

    def __init__(self, *exceptions: Type[Exception]) -> None:
        self._value = exceptions


@enum.unique
class _FromType(enum.Enum):
    OBJECT = "object"
    ARRAY = "array"
    LEAF = "leaf"


FromType = AbstractSet[_FromType]

OBJECT: Final = frozenset({_FromType.OBJECT})
ARRAY: Final = frozenset({_FromType.ARRAY})
VALUE: Final = frozenset({_FromType.LEAF})

_AUTO_KEY: Final[str] = f"{__name__}.auto"

_T_in = TypeVar("_T_in")


@dataclass(frozen=True)
class _auto(Metadata[Optional[FromType]]):
    KEY: ClassVar[str] = _AUTO_KEY
    DEFAULT: ClassVar[Optional[FromType]] = None

    @dataclass(frozen=True)
    class inner(Metadata[Optional[FromType]]):
        KEY: ClassVar[str] = _AUTO_KEY
        DEFAULT: ClassVar[Optional[FromType]] = None

        _value: FromType

        @property
        def value(self) -> FromType:
            return self._value

    @property
    def value(self) -> FromType:
        return OBJECT

    @overload
    def __call__(self, __enable: bool) -> "_auto.inner":
        ...  # pragma: no cover

    @overload
    def __call__(self, *, from_: FromType) -> "_auto.inner":
        ...  # pragma: no cover

    @overload
    def __call__(self, target: _Target) -> _Target:
        ...  # pragma: no cover

    def __call__(
        self,
        target: Optional[Union[bool, Target[_T_in]]] = None,
        from_: Optional[FromType] = None,
    ) -> Union["_auto.inner", Target[_T_in]]:
        if isinstance(target, bool):
            # First overload
            if target:
                return _auto.inner(OBJECT)
            return _auto.inner(frozenset())
        if from_ is not None:
            # Second overload
            return _auto.inner(from_)
        # Third overload
        assert target is not None
        return _auto.inner(self.value)(target)


#: Automatically create a constructor for the decorated class.
auto = _auto()



"""Deserializer to construct an instance of a class type from an object."""

import logging
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from .. import type_utils
from ..core import (
    ConstructorCore,
    ContextParameter,
    InternalConstructor,
    InternalFactory,
)
from ..data import Value
from ..errors import InternalError
from ..metadata import MetadataCollection, from_context
from ..pretty_printer import LoggableArguments, LoggableTarget
from ..types import Target
from .metadata import UnknownFieldConfig, disallow_unknown_fields
from .utils import (
    OMIT,
    ContextField,
    Field,
    FromContextField,
    PseudoField,
    SingleField,
    SkippedField,
    TypedParameter,
    VariadicField,
    add_type_hints,
    handle_keyword_pseudo_field,
    handle_positional_pseudo_field,
    skip,
)

_log = logging.getLogger(__name__)


_T_co = TypeVar("_T_co", covariant=True)


@dataclass(frozen=True)
class FromObjectConstructor(ConstructorCore[_T_co]):
    _positional: Sequence[PseudoField]
    _keyword: Mapping[str, Field]
    _var_keyword: Optional[InternalConstructor[Any]]
    # Variadic positional parameters are not used by this constructor.

    _target: Target[_T_co]
    _unknown_field_config: UnknownFieldConfig

    @staticmethod
    def new(
        metadata: MetadataCollection, factory: InternalFactory, target: Target[_T_co]
    ) -> InternalConstructor[_T_co]:
        parameters = add_type_hints(target, type_utils.get_parameters(target))

        positional: List[PseudoField] = []

        index = 0
        while (
            parameters
            and parameters[0].is_positional_only()
            and not parameters[0].is_variadic()
        ):  # pragma: no cover
            parameter = parameters.pop(0)
            positional_field = FromObjectConstructor._new_positional_field(
                metadata, factory, target, parameter
            )
            if positional_field:
                _log.debug(
                    "Created field constructor for positional parameter '%s' (%d): %s",
                    parameter.name,
                    index,
                    positional_field,
                )
                positional.append(positional_field)
            else:
                _log.debug(
                    "No constructor created for positional parameter '%s' (%d): %s",
                    parameter.name,
                    index,
                    positional_field,
                )
            index += 1

        positional_only_parameters = [
            p for p in parameters if p.is_positional_only() and not p.is_variadic()
        ]
        if positional_only_parameters:  # pragma: no cover
            raise InternalError(
                f"unexpected positional-only parameters remaining: {positional_only_parameters}"
            )

        keyword: Dict[str, Field] = {}
        var_keyword = None
        while parameters:
            parameter = parameters.pop(0)
            keyword_field = FromObjectConstructor._new_keyword_field(
                metadata, factory, target, parameter
            )
            if isinstance(keyword_field, VariadicField):
                _log.debug(
                    "Created variadic constructor for keyword parameter '%s': %s",
                    parameter.name,
                    keyword_field,
                )
                var_keyword = keyword_field.constructor
            elif keyword_field:
                _log.debug(
                    "Created field constructor for keyword parameter '%s': %s",
                    parameter.name,
                    keyword_field,
                )
                keyword[parameter.name] = keyword_field
            else:
                _log.debug(
                    "No constructor created for keyword parameter '%s': %s",
                    parameter.name,
                    keyword_field,
                )

        if parameters:  # pragma: no cover
            raise InternalError(f"unexpected parameters remaining: {parameters}")

        return factory.wrap(
            target,
            FromObjectConstructor(
                _positional=positional,
                _keyword=keyword,
                _var_keyword=var_keyword,
                _target=target,
                _unknown_field_config=disallow_unknown_fields.read(metadata, target),
            ),
        )

    @staticmethod
    def _new_positional_field(
        metadata: MetadataCollection,
        factory: InternalFactory,
        target: Target[_T_co],
        parameter: TypedParameter,
    ) -> Optional[PseudoField]:
        from_context_ = from_context.read_field(metadata, target, parameter.name)
        if skip(metadata, factory, target, parameter):
            return SkippedField(parameter.default)
        elif from_context_ is not None:
            return FromContextField.from_parameter(from_context_, parameter)
        elif parameter.is_context():
            return ContextField()
        elif parameter.is_required():
            raise factory.make_error(
                "cannot create from-object constructor for target with required "
                f"positional-only parameter: {parameter.name}"
            )
        else:
            # Positional parameter is optional; omit it.
            return None

    @staticmethod
    def _new_keyword_field(
        metadata: MetadataCollection,
        factory: InternalFactory,
        target: Target[_T_co],
        parameter: TypedParameter,
    ) -> Union[Field, VariadicField, None]:
        from_context_ = from_context.read_field(metadata, target, parameter.name)
        if skip(metadata, factory, target, parameter):
            return SkippedField(parameter.default)
        elif from_context_ is not None:
            return FromContextField.from_parameter(from_context_, parameter)
        elif parameter.is_variadic():
            # The _FromObjectConstructor doesn't use variadic positional
            # parameters.
            if parameter.is_positional():
                return None
            # The _FromObjectConstructor doesn't use variadic context
            # parameters.
            if not parameter.is_context():
                return VariadicField(
                    factory.create_constructor(
                        parameter.type_hint, frame=f'field "{parameter.name}"'
                    )
                )
            return None
        elif parameter.is_context():
            return ContextField()
        else:
            return SingleField.from_parameter(metadata, factory, target, parameter)

    def __call__(self, data: Value) -> _T_co:
        object_data = data.as_object()
        field: Field

        positional: List[object] = []
        for index, field in enumerate(self._positional):
            positional.append(
                handle_positional_pseudo_field(index, field, data.context)
            )

        keyword: Dict[str, object] = {}
        for name, field in self._keyword.items():
            if isinstance(field, SingleField):
                try:
                    object_data, element = object_data.pop(name)
                except KeyError:
                    if field.required:
                        raise data.make_error(
                            f"missing required field: {name}"
                        ) from None
                    _log.debug("Keyword parameter '%s' field omitted", name)
                    continue
                value = field.constructor(element)
                _log.debug("Keyword parameter '%s' using value: %s", name, value)
                keyword[name] = value
            else:
                value = handle_keyword_pseudo_field(name, field, data.context)
                if value is not OMIT:
                    keyword[name] = value

        if self._var_keyword:
            while object_data:
                name = next(iter(object_data))
                object_data, value_data = object_data.pop(name)
                if name in self._keyword:
                    _log.debug("Skipped shadowed variadic argument '%s'", name)
                else:
                    value = self._var_keyword(value_data)
                    _log.debug("Variadic positional parameter using value: %s", value)
                    keyword[name] = value

        unknown_fields = [
            key
            for key in object_data
            if not self._unknown_field_config.ignore_field(key)
        ]
        if unknown_fields:
            keys_str = ", ".join(f'"{key}"' for key in unknown_fields)
            raise data.make_error(f"unknown field(s): {keys_str}")

        _log.debug(
            "Instantiating %s with arguments: %s",
            LoggableTarget(self._target),
            LoggableArguments(positional, keyword),
        )
        return self._target(*positional, **keyword)  # type: ignore[call-arg]

    def get_children(self) -> Iterator[InternalConstructor[Any]]:
        for field in self._keyword.values():
            if isinstance(field, SingleField):
                yield field.constructor
        if self._var_keyword:
            yield self._var_keyword

    def requires_context(self) -> Iterable[ContextParameter]:
        for i, pseudo_field in enumerate(self._positional):  # pragma: no cover
            if isinstance(pseudo_field, FromContextField):
                yield ContextParameter(
                    key=pseudo_field.key,
                    type_=pseudo_field.type_,
                    parameter=str(i),
                    required=pseudo_field.required,
                )
        for name, field in self._keyword.items():
            if isinstance(field, FromContextField):
                yield ContextParameter(
                    key=field.key,
                    type_=field.type_,
                    parameter=f"'{name}'",
                    required=field.required,
                )



"""Deserializer to construct an instance of a class type from an array."""

import logging
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from .. import iterator_utils, type_utils
from ..core import ConstructorCore, InternalConstructor, InternalFactory
from ..data import Value
from ..errors import InternalError
from ..metadata import MetadataCollection, from_context
from ..pretty_printer import LoggableArguments, LoggableTarget
from ..types import Target
from .metadata import UnknownFieldConfig, disallow_unknown_fields
from .utils import (
    OMIT,
    ContextField,
    Field,
    FromContextField,
    PseudoField,
    SingleField,
    SkippedField,
    TypedParameter,
    VariadicField,
    add_type_hints,
    handle_keyword_pseudo_field,
    handle_positional_pseudo_field,
    skip,
)

_log = logging.getLogger(__name__)

_T_co = TypeVar("_T_co", covariant=True)


@dataclass(frozen=True)
class FromArrayConstructor(ConstructorCore[_T_co]):
    _positional: Sequence[Field]
    _var_positional: Optional[InternalConstructor[Any]]
    _keyword: Mapping[str, PseudoField]
    # Variadic keyword parameters are not used by this constructor.

    _target: Target[_T_co]
    _unknown_field_config: UnknownFieldConfig

    @staticmethod
    def new(
        metadata: MetadataCollection, factory: InternalFactory, target: Target[_T_co]
    ) -> InternalConstructor[_T_co]:
        parameters = add_type_hints(target, type_utils.get_parameters(target))

        positional: List[Field] = []
        var_positional = None

        index = 0
        while parameters and parameters[0].is_positional():
            parameter = parameters.pop(0)
            positional_field = FromArrayConstructor._new_positional_field(
                metadata, factory, target, parameter
            )
            if isinstance(positional_field, VariadicField):
                _log.debug(
                    "Created variadic constructor for positional parameter '%s' (%d): %s",
                    parameter.name,
                    index,
                    positional_field,
                )
                var_positional = positional_field.constructor
            elif positional_field:
                _log.debug(
                    "Created field constructor for positional parameter '%s' (%d): %s",
                    parameter.name,
                    index,
                    positional_field,
                )
                positional.append(positional_field)
            else:
                _log.debug(
                    "No constructor created for positional parameter '%s' (%d): %s",
                    parameter.name,
                    index,
                    positional_field,
                )
            index += 1

        positional_parameters = [p for p in parameters if p.is_positional()]
        if positional_parameters:  # pragma: no cover
            raise InternalError(
                f"unexpected positional parameters remaining: {positional_parameters}"
            )

        keyword: Dict[str, PseudoField] = {}
        while parameters:
            parameter = parameters.pop(0)
            keyword_field = FromArrayConstructor._new_keyword_field(
                metadata, factory, target, parameter
            )

            if keyword_field:
                _log.debug(
                    "Created field constructor for keyword parameter '%s': %s",
                    parameter.name,
                    keyword_field,
                )
                keyword[parameter.name] = keyword_field
            else:
                _log.debug(
                    "No constructor created for keyword parameter '%s': %s",
                    parameter.name,
                    keyword_field,
                )

        if parameters:  # pragma: no cover
            raise InternalError(f"unexpected parameters remaining: {parameters}")

        return factory.wrap(
            target,
            FromArrayConstructor(
                _positional=positional,
                _var_positional=var_positional,
                _keyword=keyword,
                _target=target,
                _unknown_field_config=disallow_unknown_fields.read(metadata, target),
            ),
        )

    @staticmethod
    def _new_positional_field(
        metadata: MetadataCollection,
        factory: InternalFactory,
        target: Target[_T_co],
        parameter: TypedParameter,
    ) -> Union[Field, VariadicField, None]:
        from_context_ = from_context.read_field(metadata, target, parameter.name)
        if skip(metadata, factory, target, parameter):
            return SkippedField(parameter.default)
        elif from_context_ is not None:
            return FromContextField.from_parameter(from_context_, parameter)
        elif parameter.is_variadic():
            # The _FromArrayConstructor doesn't use variadic context
            # parameters.
            if not parameter.is_context():
                return VariadicField(
                    factory.create_constructor(
                        parameter.type_hint, frame=f'field "{parameter.name}"'
                    )
                )
            return None
        elif parameter.is_context():
            return ContextField()
        else:
            return SingleField.from_parameter(metadata, factory, target, parameter)

    @staticmethod
    def _new_keyword_field(
        metadata: MetadataCollection,
        factory: InternalFactory,
        target: Target[_T_co],
        parameter: TypedParameter,
    ) -> Optional[PseudoField]:
        from_context_ = from_context.read_field(metadata, target, parameter.name)
        if skip(metadata, factory, target, parameter):
            return SkippedField(parameter.default)
        elif from_context_ is not None:
            return FromContextField.from_parameter(from_context_, parameter)
        elif parameter.is_variadic():
            # The _FromArrayConstructor doesn't use variadic keyword
            # parameters.
            return None
        elif parameter.is_context():
            return ContextField()
        elif parameter.is_required():
            raise factory.make_error(
                "cannot create from-array constructor for target with required "
                f"keyword-only parameter: {parameter.name}"
            )
        # Covered, despite what 'coverage' says.
        else:  # pragma: no cover
            # Keyword parameter is optional; omit it.
            return None

    def __call__(self, data: Value) -> _T_co:
        array_data = data.as_array()
        iter_data = iter(array_data)

        positional: List[object] = []
        for index, field in enumerate(self._positional):
            if isinstance(field, SingleField):
                try:
                    element = next(iter_data)
                except StopIteration:
                    if field.required:
                        raise data.make_error(
                            f"too few elements ({len(array_data)}) - "
                            f"expected at least {self.min_elements} "
                        ) from None
                    break
                value = field.constructor(element)
                _log.debug("Positional parameter %d using value: %s", index, value)
                positional.append(value)
            else:
                positional.append(
                    handle_positional_pseudo_field(index, field, data.context)
                )

        if self._var_positional:
            for element in iter_data:
                value = self._var_positional(element)
                _log.debug("Variadic positional parameter using value: %s", value)
                positional.append(value)

        if (
            not iterator_utils.is_empty(iter_data)
            and not self._unknown_field_config.ignore_element()
        ):
            raise data.make_error(
                f"too many elements ({len(array_data)}) - "
                f"expected at most {self.max_elements}"
            )

        keyword: Dict[str, object] = {}
        for name, field in self._keyword.items():
            value = handle_keyword_pseudo_field(name, field, data.context)
            if value is not OMIT:
                keyword[name] = value

        _log.debug(
            "Instantiating %s with arguments: %s",
            LoggableTarget(self._target),
            LoggableArguments(positional, keyword),
        )
        return self._target(*positional, **keyword)  # type: ignore[call-arg]

    def get_children(self) -> Iterator[InternalConstructor[Any]]:
        for field in self._positional:
            if isinstance(field, SingleField):
                yield field.constructor
        if self._var_positional:
            yield self._var_positional

    @property
    def min_elements(self) -> int:
        return sum(
            1
            for field in self._positional
            if isinstance(field, SingleField) and field.required
        )

    @property
    def max_elements(self) -> int:
        if self._var_positional:  # pragma: no cover
            raise InternalError(
                "no maximum element count for constructor with variadic parameter"
            )
        return sum(
            1
            for field in self._positional
            if isinstance(field, SingleField) and field.required
        )



"""Deserializer factory to construct an instance of a class type."""

import dataclasses
import logging
from dataclasses import dataclass
from typing import Any, Iterator, Optional, Tuple, Type, TypeVar

from typing_extensions import final

from .. import type_utils
from ..core import ConstructorCore, FactoryCore, InternalConstructor, InternalFactory
from ..data import Value
from ..errors import UnsupportedTargetTypeError, ValidationException
from ..metadata import MetadataCollection
from ..types import Target
from .from_array import FromArrayConstructor
from .from_object import FromObjectConstructor
from .from_value import FromValueConstructor
from .metadata import (
    ARRAY,
    OBJECT,
    VALUE,
    FromType,
    auto,
    disallow_unknown_fields,
    handle_exception_types,
)
from .utils import add_type_hints

__all__ = [
    "ARRAY",
    "OBJECT",
    "VALUE",
    "ClassFactory",
    "FromType",
    "auto",
    "disallow_unknown_fields",
    "handle_exception_types",
]

_log = logging.getLogger(__name__)

_T_co = TypeVar("_T_co", covariant=True)


@final
@dataclass(frozen=True)
class ClassFactory(FactoryCore):
    _metadata: MetadataCollection

    @dataclass(frozen=True)
    class _Constructor(ConstructorCore[_T_co]):
        _from_object: Optional[InternalConstructor[_T_co]]
        _from_array: Optional[InternalConstructor[_T_co]]
        _from_value: Optional[InternalConstructor[_T_co]]
        _handle_exception_types: Tuple[Type[Exception], ...]

        def __call__(self, data: "Value") -> _T_co:
            def get_constructor() -> Optional[InternalConstructor[_T_co]]:
                if data.is_object() and self._from_object:
                    return self._from_object
                if data.is_array() and self._from_array:
                    return self._from_array
                if self._from_value:
                    return self._from_value
                return None

            constructor = get_constructor()
            if not constructor:
                constructors = []
                if self._from_object:
                    constructors.append("object")
                if self._from_array:
                    constructors.append("array")
                raise data.make_error(f"expected {' or '.join(constructors)}")
            core = constructor.unwrap()
            try:
                return core(data)
            except self._handle_exception_types as e:  # pylint: disable=catching-non-exception
                raise data.make_error(  # pylint: disable=bad-exception-context
                    str(e)
                ) from e

        def get_children(self) -> Iterator[InternalConstructor[Any]]:
            if self._from_object:
                yield self._from_object
            if self._from_array:
                yield self._from_array
            if self._from_value:
                yield self._from_value

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if not hasattr(target, "__call__"):
            raise UnsupportedTargetTypeError

        try:
            add_type_hints(target, type_utils.get_parameters(target))
        except type_utils.ParameterError as e:
            raise factory.make_error(str(e))

        fromtypes = _get_fromtypes(self._metadata, target)
        if fromtypes is None:
            raise factory.make_error("class construction not enabled for target")
        if not fromtypes:
            raise factory.make_error("class construction disabled for target")

        from_object = None
        if OBJECT <= fromtypes:
            from_object = FromObjectConstructor.new(self._metadata, factory, target)

        from_array = None
        if ARRAY <= fromtypes:
            from_array = FromArrayConstructor.new(self._metadata, factory, target)

        from_value = None
        if VALUE <= fromtypes:
            from_value = FromValueConstructor.new(self._metadata, factory, target)

        return type(self)._Constructor(
            _from_object=from_object,
            _from_array=from_array,
            _from_value=from_value,
            _handle_exception_types=(
                ValidationException,
                *handle_exception_types.read(self._metadata, target),
            ),
        )


def _get_fromtypes(
    metadata: MetadataCollection, target: Target[Any]
) -> Optional[FromType]:
    from_types = auto.read(metadata, target)
    if from_types is not None:
        return from_types
    if dataclasses.is_dataclass(target) or hasattr(target, "__attrs_attrs__"):
        return OBJECT
    if _is_namedtuple(target):
        return OBJECT | ARRAY
    return None


def _is_namedtuple(target: Target[_T_co]) -> bool:
    if getattr(target, "__bases__", None) != (tuple,):
        return False
    fields = getattr(target, "_fields", None)
    return isinstance(fields, tuple) and all(isinstance(field, str) for field in fields)
